/* 
 *  Author: Chris Lord <chris@openedhand.com>
 *
 *  Copyright (c) 2007 OpenedHand Ltd - http://www.openedhand.com/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */

#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <libjana-gtk/jana-gtk-tree-layout.h>
#include <libjana/jana-utils.h>
#include "jana-gtk-day-view.h"

G_DEFINE_TYPE (JanaGtkDayView, jana_gtk_day_view, GTK_TYPE_EVENT_BOX)

#define DAY_VIEW_PRIVATE(o) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((o), JANA_GTK_TYPE_DAY_VIEW, \
	JanaGtkDayViewPrivate))

typedef struct _JanaGtkDayViewPrivate JanaGtkDayViewPrivate;

struct _JanaGtkDayViewPrivate
{
	GtkWidget *alignment;
	GtkWidget *layout;
	GtkCellRenderer *event_renderer;
	GtkAllocation allocation;

	PangoLayout **time_layouts;
	PangoLayout **day_layouts;
	PangoLayout *week_layout;
	guint col0_width;
	guint row0_height;

	JanaDuration *range;
	gint visible_days;
	gint cell_minutes;

	JanaTime *day;
	JanaDuration *selection;
	guint cells;
	gchar *style_hint;
	guint spacing;
};

enum {
	PROP_RANGE = 1,
	PROP_SELECTION,
	PROP_CELLS,
	PROP_SPACING,
	PROP_STYLE_HINT,
};

enum {
	SELECTION_CHANGED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static void relayout (JanaGtkDayView *self);

static void
create_layouts (JanaGtkDayView *self)
{
	gint x, text_width, text_height;
	gchar week_string[4];
	JanaTime *time, *end;
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	if ((!jana_duration_valid (priv->range)) || (priv->cells == 0)) return;
	
	/* Work out how many minutes there are per cell */
	/* TODO: This is an extremely slow way of working
	 *       this out, but this isn't a speed critical
	 *       section of code. Resolve in the future.
	 */
	time = jana_time_duplicate (priv->range->start);
	end = jana_time_duplicate (time);
	jana_time_set_hours (end, jana_time_get_hours (priv->range->end));
	jana_time_set_minutes (end, jana_time_get_minutes (priv->range->end));
	if (jana_utils_time_compare (time, end, FALSE) >= 0)
		jana_time_set_day (end, jana_time_get_day (end) + 1);
	for (priv->cell_minutes = 0;
	     jana_utils_time_compare (time, end, FALSE) < 0;
	     jana_time_set_minutes (time, jana_time_get_minutes (time) +
				    priv->cells),
	     priv->cell_minutes ++) {
		if (jana_time_get_day (time) !=
		    jana_time_get_day (priv->range->start))
			break;
	}
	g_object_unref (end);
	g_object_unref (time);

	priv->row0_height = 0;
	priv->day_layouts = g_slice_alloc (
		sizeof (PangoLayout *) * priv->visible_days);
	time = jana_time_duplicate (priv->range->start);
	for (x = 0; x < priv->visible_days; x++) {
		GDateWeekday day;
		gchar day_markup[32];

		day = jana_utils_time_day_of_week (time);
		snprintf (day_markup, 32, "%c\n<small>%d/%d</small>",
			nl_langinfo (ABDAY_1 + (day % 7))[0],
			jana_time_get_day (time),
			jana_time_get_month (time));
		
		priv->day_layouts[x] = gtk_widget_create_pango_layout (
			GTK_WIDGET (self), NULL);
		pango_layout_set_markup (priv->day_layouts[x], day_markup, -1);
		pango_layout_set_alignment (
			priv->day_layouts[x], PANGO_ALIGN_CENTER);
		
		pango_layout_get_pixel_size (
			priv->day_layouts[x], &text_width, &text_height);
		
		/* Set the height of the first row */
		priv->row0_height = MAX (priv->row0_height, text_height);
		jana_time_set_day (time, jana_time_get_day (time) + 1);
	}
	g_object_unref (time);

	priv->col0_width = 0;
	priv->time_layouts = g_slice_alloc (
		sizeof (PangoLayout *) * priv->cells);
	time = jana_time_duplicate (priv->range->start);
	for (x = 0; x < priv->cells; x++) {
		gchar time_string[6];

		snprintf (time_string, 6, "%02d:%02d",
			jana_time_get_hours (time),
			jana_time_get_minutes (time));
		
		priv->time_layouts[x] = gtk_widget_create_pango_layout (
			GTK_WIDGET (self), time_string);
		
		pango_layout_get_pixel_size (
			priv->time_layouts[x], &text_width, &text_height);
		
		/* Set the width of the first column */
		priv->col0_width = MAX (priv->col0_width, text_width);
		jana_time_set_minutes (time,
			jana_time_get_minutes (time) + priv->cell_minutes);
	}
	g_object_unref (time);

	snprintf (week_string, 4, "W%02d",
		jana_utils_time_week_of_year (priv->range->start, TRUE));
	priv->week_layout = gtk_widget_create_pango_layout (
		GTK_WIDGET (self), NULL);
	pango_layout_set_text (priv->week_layout, week_string, -1);
	pango_layout_get_pixel_size (priv->week_layout,
		&text_width, &text_height);
	
	/* Check the widths/heights, just in case */
	priv->col0_width = MAX (priv->col0_width, text_width);
	priv->row0_height = MAX (priv->row0_height, text_height);

	/* Set the alignment on the tree layout */
	gtk_alignment_set_padding (GTK_ALIGNMENT (priv->alignment),
		priv->row0_height + (priv->spacing * 2), 0,
		priv->col0_width + (priv->spacing * 2), 0);
}

static void
free_layouts (JanaGtkDayView *self)
{
	gint i;
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	if (priv->day_layouts) {
		for (i = 0; i < priv->visible_days; i++)
			g_object_unref (priv->day_layouts[i]);
		g_slice_free1 (sizeof (PangoLayout*) * priv->visible_days,
			priv->day_layouts);
		priv->day_layouts = NULL;
	}

	if (priv->time_layouts) {
		for (i = 0; i < priv->cells; i++)
			g_object_unref (priv->time_layouts[i]);
		g_slice_free1 (sizeof (PangoLayout*) * priv->cells,
			priv->time_layouts);
	}

	if (priv->week_layout) {
		g_object_unref (priv->week_layout);
		priv->week_layout = NULL;
	}
	
}

static void
jana_gtk_day_view_get_property (GObject *object, guint property_id,
				GValue *value, GParamSpec *pspec)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (object);

	switch (property_id) {
	    case PROP_RANGE :
		g_value_set_boxed (value, priv->range);
		break;
	    case PROP_SPACING :
		g_value_set_uint (value, priv->spacing);
		break;
	    case PROP_CELLS :
		g_value_set_uint (value, priv->cells);
		break;
	    case PROP_STYLE_HINT :
		g_value_set_string (value, priv->style_hint);
		break;
	    case PROP_SELECTION :
		g_value_set_boxed (value, priv->selection);
		break;
	    default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
jana_gtk_day_view_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (object);

	switch (property_id) {
	    case PROP_RANGE :
		if (priv->range) jana_duration_free (priv->range);
		priv->range = (JanaDuration *)g_value_dup_boxed (value);
		if (priv->range) {
			JanaTime *time;
			
			jana_time_set_isdate (priv->range->start, FALSE);
			jana_time_set_isdate (priv->range->end, FALSE);
			
			/* Count visible days */
			for (priv->visible_days = 0,
			     time = jana_time_duplicate (priv->range->start);
			     jana_utils_time_compare (
				time, priv->range->end, TRUE) < 0;
			     jana_time_set_day (
				time, jana_time_get_day (time) + 1),
			     priv->visible_days ++);

			g_object_unref (time);
		}
		if (priv->selection) {
			jana_duration_free (priv->selection);
			priv->selection = NULL;
			g_signal_emit (object,
				signals[SELECTION_CHANGED], 0, NULL);
		}
		free_layouts (JANA_GTK_DAY_VIEW (object));
		create_layouts (JANA_GTK_DAY_VIEW (object));
		relayout (JANA_GTK_DAY_VIEW (object));
		gtk_widget_queue_draw (GTK_WIDGET (object));
		break;
	    case PROP_SPACING :
		priv->spacing = g_value_get_uint (value);
		gtk_alignment_set_padding (GTK_ALIGNMENT (priv->alignment),
			priv->row0_height + (priv->spacing * 2), 0,
			priv->col0_width + (priv->spacing * 2), 0);
		gtk_widget_queue_draw (GTK_WIDGET (object));
		break;
	    case PROP_CELLS :
		priv->cells = g_value_get_uint (value);
		free_layouts (JANA_GTK_DAY_VIEW (object));
		create_layouts (JANA_GTK_DAY_VIEW (object));
		relayout (JANA_GTK_DAY_VIEW (object));
		gtk_widget_queue_draw (GTK_WIDGET (object));
		break;
	    case PROP_STYLE_HINT :
		if (priv->style_hint) {
			g_free (priv->style_hint);
			priv->style_hint = NULL;
		}
		priv->style_hint = g_value_dup_string (value);
		gtk_widget_queue_draw (GTK_WIDGET (object));
		break;
	    case PROP_SELECTION :
		if (priv->selection) jana_duration_free (priv->selection);
		priv->selection = (JanaDuration *)g_value_dup_boxed (value);
		g_signal_emit (object, signals[SELECTION_CHANGED],
			0, priv->selection);
		gtk_widget_queue_draw (GTK_WIDGET (object));
		break;
	    default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
jana_gtk_day_view_dispose (GObject *object)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (object);

	if (priv->event_renderer) {
		g_object_unref (priv->event_renderer);
		priv->event_renderer = NULL;
	}
	
	if (G_OBJECT_CLASS (jana_gtk_day_view_parent_class)->dispose)
		G_OBJECT_CLASS (jana_gtk_day_view_parent_class)->
			dispose (object);
}

static void
jana_gtk_day_view_finalize (GObject *object)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (object);

	if (priv->range) {
		jana_duration_free (priv->range);
		priv->range = NULL;
	}
	
	if (priv->selection) {
		jana_duration_free (priv->selection);
		priv->selection = NULL;
	}
	
	if (priv->style_hint) {
		g_free (priv->style_hint);
		priv->style_hint = NULL;
	}
	
	free_layouts (JANA_GTK_DAY_VIEW (object));

	G_OBJECT_CLASS (jana_gtk_day_view_parent_class)->finalize (object);
}

static gboolean
jana_gtk_day_view_expose_event (GtkWidget *widget, GdkEventExpose *event)
{
	gint x, y, box_x, row_height, col_width, text_width, text_height,
		selection_direction = 0;
	GtkStyle *dark_style, *style;
	JanaTime *time;

	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (widget);
	
	if ((!jana_duration_valid (priv->range)) || (priv->cells == 0))
		return FALSE;
	
	if (priv->selection) selection_direction = jana_utils_time_compare (
		priv->selection->start, priv->selection->end, FALSE);
	
	/* Draw background */
	box_x = 0;
	row_height = (widget->allocation.height - priv->row0_height -
		(priv->spacing * 2)) / priv->cells;
	col_width = (widget->allocation.width - priv->col0_width -
		(priv->spacing * 2)) / priv->visible_days;

	dark_style = gtk_style_copy (widget->style);
	dark_style->bg[GTK_STATE_NORMAL] = dark_style->bg[GTK_STATE_ACTIVE];
	dark_style->fg[GTK_STATE_NORMAL] = dark_style->fg[GTK_STATE_ACTIVE];
	dark_style = gtk_style_attach (dark_style, widget->window);

	time = jana_time_duplicate (priv->range->start);
	jana_time_set_day (time, jana_time_get_day (time) - 1);
	for (x = 0; x < priv->visible_days + 1; x++) {
		gint box_y = 0, box_width, box_height, day;
		
		if (x == 0) {
			box_width = priv->col0_width + (priv->spacing * 2);
		} else {
			box_width = col_width;
		}
		
		day = jana_time_get_day (time);
		
		for (y = 0; y < priv->cells + 1; y++) {
			GtkShadowType shadow = GTK_SHADOW_IN;
			GtkStateType state = GTK_STATE_NORMAL;
			PangoLayout *layout = NULL;

			style = widget->style;
			
			if (y == 0) {
				box_height = priv->row0_height +
					(priv->spacing * 2);
				if (x == 0) layout = priv->week_layout;
				else layout = priv->day_layouts[x - 1];
			} else {
				box_height = row_height;
				if (x == 0) layout = priv->time_layouts[y - 1];
			}
		

			if ((x != 0) && (y != 0)) {
				/* Possible selection */
				if (priv->selection) {
					if ((selection_direction < 0) &&
					    (jana_utils_time_compare (time,
					     priv->selection->start,
					     FALSE) >= 0) &&
					    (jana_utils_time_compare (time,
					     priv->selection->end,
					     FALSE) < 0)) {
						state = GTK_STATE_SELECTED;
					} else
					if ((selection_direction > 0) &&
					    (jana_utils_time_compare (time,
					     priv->selection->end,
					     FALSE) >= 0) &&
					    (jana_utils_time_compare (time,
					     priv->selection->start,
					     FALSE) < 0)) {
						state = GTK_STATE_SELECTED;
					}
				}

				jana_time_set_minutes (time,
					jana_time_get_minutes (time) +
					priv->cell_minutes);
			}
			
			if ((x == 0) || (y == 0)) {
				shadow = GTK_SHADOW_OUT;
				style = dark_style;
			}
			
			gtk_paint_box (style, widget->window,
				state, shadow,
				&event->area, widget, priv->style_hint,
				box_x, box_y,
				box_width, box_height);
			
			if (layout) {
				pango_layout_get_pixel_size (
					layout, &text_width, &text_height);
			
				gtk_paint_layout (style, widget->window,
					state, FALSE, &event->area,
					widget, priv->style_hint,
					box_x + (box_width / 2) -
						  (text_width / 2),
					box_y + (box_height / 2) -
						  (text_height / 2),
					layout);
			}
			box_y += box_height;
		}
		jana_time_set_hours (time,
			jana_time_get_hours (priv->range->start));
		jana_time_set_minutes (time,
			jana_time_get_minutes (priv->range->start));
		jana_time_set_day (time, day + 1);

		box_x += box_width;
	}
	gtk_style_detach (dark_style);
	
	GTK_WIDGET_CLASS (jana_gtk_day_view_parent_class)->
		expose_event (widget, event);
	
	g_object_unref (time);
	
	return FALSE;
}

static void
relayout (JanaGtkDayView *self)
{
	GList *cell, *cells;
	gint cell_width, min_time, max_time, alloc_height;
	gfloat min_per_pixel;

	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	if ((!jana_duration_valid (priv->range)) || (priv->cells == 0))
		return;

	cell_width = priv->layout->allocation.width / priv->visible_days;
	min_time = (jana_time_get_hours (priv->range->start) * 60) +
		jana_time_get_minutes (priv->range->start);
	max_time = ((jana_time_get_hours (priv->range->end) ?
		     jana_time_get_hours (priv->range->end) : 24) * 60) +
		jana_time_get_minutes (priv->range->end);
	
	/* TODO: Use integer math for this */
	alloc_height = priv->layout->allocation.height -
		(priv->layout->allocation.height % priv->cells);
	min_per_pixel = (gfloat)alloc_height / (gfloat)(max_time - min_time);
	
	cells = jana_gtk_tree_layout_get_cells (
		JANA_GTK_TREE_LAYOUT (priv->layout));
	
	for (cell = g_list_last (cells); cell; cell = cell->prev) {
		JanaGtkTreeLayoutCellInfo *info =
			(JanaGtkTreeLayoutCellInfo *)cell->data;
		GtkTreeModel *model;
		GtkTreePath *path;
		GtkTreeIter iter;
		JanaTime *start, *end, *time;
		gint day;

		/* Work out what day the cell is in */
		model = gtk_tree_row_reference_get_model (info->row);
		path = gtk_tree_row_reference_get_path (info->row);
		gtk_tree_model_get_iter (model, &iter, path);
		gtk_tree_path_free (path);
	
		gtk_tree_model_get (model, &iter,
			JANA_GTK_EVENT_STORE_COL_START, &start,
			JANA_GTK_EVENT_STORE_COL_END, &end, -1);
		
		if (!start) {
			if (end) g_object_unref (end);
			continue;
		}
		if (!end) {
			end = jana_time_duplicate (start);
			jana_time_set_minutes (
				end, jana_time_get_minutes (end) + 30);
		}
		
		/* Get to the correct day */
		for (time = jana_time_duplicate (priv->range->start), day = 0;
		     jana_utils_time_compare (time, start, TRUE) < 0;
		     jana_time_set_day (time, jana_time_get_day (time) + 1),
		     day ++);
		if (jana_utils_time_compare (time, start, TRUE) == 0) {
			GList *prev_cell;
			JanaGtkTreeLayoutCellInfo *ovl_info = NULL;
			gint x, y, width, height, minutes;
			
			x = (cell_width * day) + priv->spacing;
			width = cell_width - (priv->spacing * 2);
			
			minutes = (jana_time_get_hours (start) * 60) +
				jana_time_get_minutes (start);
			minutes -= min_time;
			
			y = min_per_pixel * minutes;

			minutes = (jana_time_get_hours (end) * 60) +
				jana_time_get_minutes (end);
			minutes -= min_time;
			
			height = (min_per_pixel * minutes) - y;
			
			if (y < 0) {
				height += y;
				y = 0;
			}
			
			if ((width <= 0) || (height <= 0) ||
			    (y > alloc_height)) {
				/* Hide off-screen events */
				x = y = width = height = 0;
			} else if (cell->next) {
				/* Check if we've overlapped with any previous
				 * events in this day.
				 */
				for (prev_cell = cell->next; prev_cell;
				     prev_cell = prev_cell->next) {
					JanaGtkTreeLayoutCellInfo *prev_info =
						(JanaGtkTreeLayoutCellInfo *)
						prev_cell->data;
					     
					/* Check if info is on a previous day */
					if (prev_info->x < x) break;
					
					/* Check if event is hidden */
					if ((prev_info->width == 0) ||
					    (prev_info->height == 0))
						continue;
					     
					/* Check if we're overlapping */
					if (y < (prev_info->y +
					    prev_info->height)) {
						ovl_info = prev_info;
						break;
					}
				}
			}
			
			/* Resize on overlap */
			if (ovl_info) {
				/* 1/2 event width is enough room to squeeze
				 * it in without resizing the last event.
				 */
				if (ovl_info->x >= (x + width/2)) {
					width = ovl_info->x - x -
						priv->spacing;
				} else {
					/* Reduce the size of the last event
					 * and shuffle this event along
					 * slightly.
					 */
					width = (ovl_info->width * 2)/3;
					x = ovl_info->x + (ovl_info->width -
							    width);
					ovl_info->width = width;

					/* Resize overlapped event */
					jana_gtk_tree_layout_move_cell (
						JANA_GTK_TREE_LAYOUT (
							priv->layout),
						ovl_info->row, ovl_info->x,
						ovl_info->y, ovl_info->width,
						ovl_info->height);
				}
			}

			jana_gtk_tree_layout_move_cell (
				JANA_GTK_TREE_LAYOUT (priv->layout),
				info->row, x, y, width, height);
		} else {
			/* Hide out-of-range events */
			jana_gtk_tree_layout_move_cell (
				JANA_GTK_TREE_LAYOUT (priv->layout),
				info->row, 0, 0, 0, 0);
		}
		
		g_object_unref (time);
		g_object_unref (start);
		g_object_unref (end);
	}
}

static void
size_request_cb (GtkWidget *widget, GtkRequisition *requisition,
		 JanaGtkDayView *self)
{
	requisition->width = 0;
	requisition->height = 0;
}

static void
size_allocate_cb (GtkWidget *widget, GtkAllocation *allocation,
		  JanaGtkDayView *self)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);

	if ((allocation->width != priv->allocation.width) ||
	    (allocation->height != priv->allocation.height)) {
		priv->allocation.width = allocation->width;
		priv->allocation.height = allocation->height;
		relayout (self);
	}
}

static void
start_selection (JanaGtkDayView *self, gint x, gint y)
{
	JanaTime *start, *end;
	gint alloc_height;

	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	alloc_height = priv->layout->allocation.height -
		(priv->layout->allocation.height % priv->cells);
	x = x + 1;
	y = MIN (y, alloc_height - 1);

	start = jana_time_duplicate (priv->range->start);
	jana_time_set_day (start, jana_time_get_day (start) +
		(x * priv->visible_days) /
		priv->layout->allocation.width);
	jana_time_set_minutes (start, jana_time_get_minutes (start) +
		(((y * priv->cells) /
		  alloc_height) * priv->cell_minutes));
	end = jana_time_duplicate (start);
	jana_time_set_minutes (end, jana_time_get_minutes (end) +
		priv->cell_minutes);
	
	if (priv->selection) jana_duration_free (priv->selection);
	priv->selection = jana_duration_new (start, end);
	
	g_object_unref (start);
	g_object_unref (end);
}

static void
end_selection (JanaGtkDayView *self, gint x, gint y)
{
	gint alloc_height;
	
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	alloc_height = priv->layout->allocation.height -
		(priv->layout->allocation.height % priv->cells);
	x = MAX (MIN (x + 1, priv->layout->allocation.width), 0);
	y = MAX (MIN (y + 1, alloc_height), 0);

	jana_time_set_day (priv->selection->end,
		jana_time_get_day (priv->range->start) +
		(x * priv->visible_days) /
		priv->layout->allocation.width);
	jana_time_set_hours (priv->selection->end,
		jana_time_get_hours (priv->range->start));
	jana_time_set_minutes (priv->selection->end,
		jana_time_get_minutes (priv->range->start) +
		((((y * priv->cells) /
		  alloc_height) + 1) *
		 priv->cell_minutes));
	
	if (jana_utils_time_compare (
	    priv->selection->end, priv->selection->start, FALSE) <= 0)
		jana_time_set_minutes (priv->selection->end,
			jana_time_get_minutes (priv->selection->end) -
			priv->cell_minutes);
}

static gboolean
button_press_event_cb (GtkWidget *layout, GdkEventButton *event,
		       JanaGtkDayView *self)
{
	GList *selection;
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	if ((event->button != 1) || (!jana_duration_valid (priv->range)) ||
	    (priv->cells == 0)) return FALSE;
	
	selection = jana_gtk_tree_layout_get_selection (
		JANA_GTK_TREE_LAYOUT (layout));
	
	if (!selection) {
		start_selection (self, event->x, event->y);
		gtk_widget_queue_draw (GTK_WIDGET (self));
	} else {
		if (priv->selection) {
			jana_duration_free (priv->selection);
			priv->selection = NULL;
			gtk_widget_queue_draw (GTK_WIDGET (self));
		}
		g_list_free (selection);
	}
	
	return FALSE;
}

static gboolean
button_release_event_cb (GtkWidget *layout, GdkEventButton *event,
			 JanaGtkDayView *self)
{
	gint result;
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);

	if ((event->button != 1) || (!priv->selection)) return FALSE;
	
	end_selection (self, event->x, event->y);

	/* Swap start/end if end is before start */
	result = jana_utils_time_compare (priv->selection->end,
		priv->selection->start, FALSE);
	if (result < 0) {
		JanaTime *start;
		start = priv->selection->start;
		priv->selection->start = priv->selection->end;
		priv->selection->end = start;
	} else if (result == 0) {
		jana_duration_free (priv->selection);
		priv->selection = NULL;
	}

	gtk_widget_queue_draw (GTK_WIDGET (self));
	g_signal_emit (self, signals[SELECTION_CHANGED], 0, priv->selection);
	
	return FALSE;
}

static gboolean
motion_notify_event_cb (GtkWidget *layout, GdkEventMotion *event,
			JanaGtkDayView *self)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);

	if ((event->state & GDK_BUTTON1_MASK) && (priv->selection)) {
		end_selection (self, event->x, event->y);
		gtk_widget_queue_draw (GTK_WIDGET (self));
	}
	
	/* Request more motion events */
#if GTK_CHECK_VERSION(2, 12, 0)
	gdk_event_request_motions (event);
#else
	gdk_window_get_pointer (layout->window, NULL, NULL, NULL);
#endif
	
	return FALSE;
}

static void
jana_gtk_day_view_class_init (JanaGtkDayViewClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

	g_type_class_add_private (klass, sizeof (JanaGtkDayViewPrivate));

	object_class->get_property = jana_gtk_day_view_get_property;
	object_class->set_property = jana_gtk_day_view_set_property;
	object_class->dispose = jana_gtk_day_view_dispose;
	object_class->finalize = jana_gtk_day_view_finalize;
	
	widget_class->expose_event = jana_gtk_day_view_expose_event;

	g_object_class_install_property (
		object_class,
		PROP_RANGE,
		g_param_spec_boxed (
			"range",
			"JanaDuration *",
			"The range shown by this day-view, as a JanaDuration.",
			JANA_TYPE_DURATION,
			G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_CELLS,
		g_param_spec_uint (
			"cells",
			"guint",
			"Amount of cells partition time into, per day.",
			0, G_MAXUINT, 0,
			G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_SPACING,
		g_param_spec_uint (
			"spacing",
			"guint",
			"Spacing to use inside day boxes.",
			0, G_MAXUINT, 2,
			G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_STYLE_HINT,
		g_param_spec_string (
			"style_hint",
			"gchar *",
			"The style hint to use when drawing.",
			NULL,
			G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_SELECTION,
		g_param_spec_boxed (
			"selection",
			"JanaDuration *",
			"The currently selected range, as a JanaDuration.",
			JANA_TYPE_DURATION,
			G_PARAM_READWRITE));

	signals[SELECTION_CHANGED] =
		g_signal_new ("selection_changed",
			G_OBJECT_CLASS_TYPE (object_class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (JanaGtkDayViewClass,
					 selection_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOXED,
			G_TYPE_NONE, 1, JANA_TYPE_DURATION);
}

static gint
sort_cells_cb (JanaGtkTreeLayoutCellInfo *info_a,
	       JanaGtkTreeLayoutCellInfo *info_b)
{
	GtkTreeModel *model;
	GtkTreePath *path;
	GtkTreeIter iter;
	JanaTime *start_a, *end_a, *start_b, *end_b;
	gchar *summary_a, *summary_b;
	gint result = 0;

	model = gtk_tree_row_reference_get_model (info_a->row);
	path = gtk_tree_row_reference_get_path (info_a->row);
	gtk_tree_model_get_iter (model, &iter, path);
	gtk_tree_path_free (path);
	
	gtk_tree_model_get (model, &iter,
		JANA_GTK_EVENT_STORE_COL_START, &start_a,
		JANA_GTK_EVENT_STORE_COL_END, &end_a,
		JANA_GTK_EVENT_STORE_COL_SUMMARY, &summary_a, -1);

	model = gtk_tree_row_reference_get_model (info_b->row);
	path = gtk_tree_row_reference_get_path (info_b->row);
	gtk_tree_model_get_iter (model, &iter, path);
	gtk_tree_path_free (path);
	
	gtk_tree_model_get (model, &iter,
		JANA_GTK_EVENT_STORE_COL_START, &start_b,
		JANA_GTK_EVENT_STORE_COL_END, &end_b,
		JANA_GTK_EVENT_STORE_COL_SUMMARY, &summary_b, -1);
	
	if (start_a && start_b)
		result = jana_utils_time_compare (start_b, start_a, FALSE);
	if (result == 0) {
		if (end_a && end_b)
			result = jana_utils_time_compare (end_a, end_b, FALSE);
		if ((result == 0) && summary_a && summary_b)
			result = g_utf8_collate (summary_b, summary_a);
	}
	
	g_free (summary_a);
	g_free (summary_b);
	if (start_a) g_object_unref (start_a);
	if (start_b) g_object_unref (start_b);
	if (end_a) g_object_unref (end_a);
	if (end_b) g_object_unref (end_b);
	
	return result;
}

static void
style_set_cb (GtkWidget *widget, GtkStyle *previous_style, JanaGtkDayView *self)
{
	static gboolean skip_next = FALSE;
	if (!skip_next) {
		skip_next = TRUE;
		gtk_widget_modify_bg (widget, GTK_STATE_NORMAL,
			&widget->style->base[GTK_STATE_NORMAL]);
		skip_next = FALSE;
	}
}

static void
jana_gtk_day_view_init (JanaGtkDayView *self)
{
	static gboolean locale_set = FALSE;
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);

	if (!locale_set) {
		setlocale (LC_TIME, "");
		locale_set = TRUE;
	}
	
	gtk_widget_set_app_paintable (GTK_WIDGET (self), TRUE);

	priv->spacing = 2;
	priv->alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
	priv->layout = jana_gtk_tree_layout_new ();
	g_object_set (G_OBJECT (priv->layout),
		"sort_cb", sort_cells_cb,
		"select_mode", GTK_SELECTION_SINGLE,
		NULL);
	gtk_widget_add_events (priv->layout, GDK_POINTER_MOTION_HINT_MASK);
	
	gtk_container_add (GTK_CONTAINER (priv->alignment), priv->layout);
	gtk_container_add (GTK_CONTAINER (self), priv->alignment);
	gtk_widget_show (priv->alignment);
	gtk_widget_show (priv->layout);
	
	priv->event_renderer = g_object_ref_sink (
		jana_gtk_cell_renderer_event_new ());
	g_object_set (G_OBJECT (priv->event_renderer),
		"draw_text", FALSE, NULL);
	
	g_signal_connect (priv->layout, "size-request",
		G_CALLBACK (size_request_cb), self);
	g_signal_connect (priv->layout, "size-allocate",
		G_CALLBACK (size_allocate_cb), self);
	g_signal_connect_after (priv->layout, "button-press-event",
		G_CALLBACK (button_press_event_cb), self);
	g_signal_connect (priv->layout, "button-release-event",
		G_CALLBACK (button_release_event_cb), self);
	g_signal_connect (priv->layout, "motion-notify-event",
		G_CALLBACK (motion_notify_event_cb), self);
	
	g_signal_connect (self, "style-set",
		G_CALLBACK (style_set_cb), self);
}

GtkWidget *
jana_gtk_day_view_new (JanaDuration *range, guint cells)
{
	return GTK_WIDGET (g_object_new (JANA_GTK_TYPE_DAY_VIEW,
		"range", range, "cells", cells, NULL));
}

static void
row_changed_cb (GtkTreeModel *tree_model, GtkTreePath *path,
		GtkTreeIter *iter, JanaGtkDayView *self)
{
	relayout (self);
}

static void
row_inserted_cb (GtkTreeModel *tree_model, GtkTreePath *path,
		 GtkTreeIter *iter, JanaGtkDayView *self)
{
	GtkTreeRowReference *row;
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	/* Add row to layout */
	row = gtk_tree_row_reference_new (tree_model, path);
	
	jana_gtk_tree_layout_add_cell (JANA_GTK_TREE_LAYOUT (priv->layout),
		row, 0, 0, 0, 0, priv->event_renderer,
		"uid", JANA_GTK_EVENT_STORE_COL_UID,
		"summary", JANA_GTK_EVENT_STORE_COL_SUMMARY,
		"description", JANA_GTK_EVENT_STORE_COL_DESCRIPTION,
		"start", JANA_GTK_EVENT_STORE_COL_START,
		"end", JANA_GTK_EVENT_STORE_COL_END,
		"first_instance", JANA_GTK_EVENT_STORE_COL_FIRST_INSTANCE,
		"last_instance", JANA_GTK_EVENT_STORE_COL_LAST_INSTANCE,
		"has_recurrences", JANA_GTK_EVENT_STORE_COL_HAS_RECURRENCES,
		"has_alarm", JANA_GTK_EVENT_STORE_COL_HAS_ALARM,
		NULL);
	
	gtk_tree_row_reference_free (row);
	
	relayout (self);
}

static void
row_deleted_cb (GtkTreeModel *tree_model, GtkTreePath *path,
		JanaGtkDayView *self)
{
	relayout (self);
}

static void
insert_rows (JanaGtkDayView *self, JanaGtkEventStore *store)
{
	GtkTreeIter iter;
	
	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (store), &iter)) do {
		GtkTreePath *path = gtk_tree_model_get_path (
			(GtkTreeModel *)store, &iter);
		row_inserted_cb ((GtkTreeModel *)store, path, &iter, self);
		gtk_tree_path_free (path);
	} while (gtk_tree_model_iter_next ((GtkTreeModel *)store, &iter));
}

static void
remove_rows (JanaGtkDayView *self, JanaGtkEventStore *store)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	GList *cell, *cells = jana_gtk_tree_layout_get_cells (
		JANA_GTK_TREE_LAYOUT (priv->layout));
	
	for (cell = cells; cell; cell = cell->next) {
		JanaGtkTreeLayoutCellInfo *info =
			(JanaGtkTreeLayoutCellInfo *)cell->data;
		if (gtk_tree_row_reference_get_model (info->row) ==
		    (GtkTreeModel *)store)
			jana_gtk_tree_layout_remove_cell (
				(JanaGtkTreeLayout *)self, info->row);
	}
	
	g_list_free (cells);
}

void
jana_gtk_day_view_add_store (JanaGtkDayView *self, JanaGtkEventStore *store)
{
	insert_rows (self, store);
	
	g_signal_connect (store, "row-inserted",
		G_CALLBACK (row_inserted_cb), self);
	g_signal_connect_after (store, "row-changed",
		G_CALLBACK (row_changed_cb), self);
	g_signal_connect_after (store, "row-deleted",
		G_CALLBACK (row_deleted_cb), self);
}

void
jana_gtk_day_view_remove_store (JanaGtkDayView *self,
				  JanaGtkEventStore *store)
{
	g_signal_handlers_disconnect_by_func (store, row_inserted_cb, self);
	g_signal_handlers_disconnect_by_func (store, row_changed_cb, self);
	g_signal_handlers_disconnect_by_func (store, row_deleted_cb, self);
	
	remove_rows (self, store);
}

JanaGtkCellRendererEvent *
jana_gtk_day_view_get_cell_renderer (JanaGtkDayView *self)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	return (JanaGtkCellRendererEvent *)priv->event_renderer;
}

void
jana_gtk_day_view_set_range (JanaGtkDayView *self, JanaDuration *range)
{
	g_object_set (G_OBJECT (self), "range", range, NULL);
}

JanaDuration *
jana_gtk_day_view_get_range (JanaGtkDayView *self)
{
	JanaDuration *range;
	
	g_object_get (G_OBJECT (self), "range", &range, NULL);
	
	return range;
}

JanaDuration *
jana_gtk_day_view_get_selection (JanaGtkDayView *self)
{
	JanaDuration *selection;
	
	g_object_get (G_OBJECT (self), "selection", &selection, NULL);
	
	return selection;
}

void
jana_gtk_day_view_set_spacing (JanaGtkDayView *self, guint spacing)
{
	g_object_set (G_OBJECT (self), "spacing", spacing, NULL);
}

guint
jana_gtk_day_view_get_spacing (JanaGtkDayView *self)
{
	guint spacing;
	
	g_object_get (G_OBJECT (self), "spacing", &spacing, NULL);
	
	return spacing;
}

JanaGtkTreeLayout *
jana_gtk_day_view_get_layout (JanaGtkDayView *self)
{
	JanaGtkDayViewPrivate *priv = DAY_VIEW_PRIVATE (self);
	
	return JANA_GTK_TREE_LAYOUT (priv->layout);
}
