/*
 *  openmoko-dates -- OpenMoko Calendar Application
 *
 *  Authored by Chris Lord <chris@openedhand.com>
 *
 *  Copyright (C) 2006-2007 OpenMoko Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser Public License as published by
 *  the Free Software Foundation; version 2 of the license.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *
 *  Current Version: $Rev$ ($Date$) [$Author$]
 */

#include <string.h>
#include <gtk/gtk.h>
#include <libjana/jana.h>
#include <libjana-ecal/jana-ecal.h>
#include <libjana-gtk/jana-gtk.h>
#include <libmokoui2/moko-finger-scroll.h>
#include <libmokoui2/moko-search-bar.h>
#include "dates.h"
#include "dates-week.h"
#include "dates-utils.h"

static gboolean hidden = TRUE;


static void
refresh_view (DatesData *data)
{
	JanaDuration *range;

	range = jana_duration_new (data->start, data->end);
	jana_gtk_day_view_set_range (JANA_GTK_DAY_VIEW (data->week_view),
		range);
	jana_duration_free (range);
	
	dates_utils_refresh_query (data);
}

static void
page_shown (DatesData *data)
{
	/* Set the range on the day-view and queries */
	
	/* If the last visible week-view or list-view times are in range,
	 * use them.
	 */
	if (data->week_time && (jana_utils_time_compare (
	     data->list_time, data->start, TRUE) > 0) &&
	    (jana_utils_time_compare (
	     data->list_time, data->end, TRUE) < 0)) {
		g_object_unref (data->start);
		data->start = jana_time_duplicate (data->week_time);
	} else if (data->list_time && (jana_utils_time_compare (
	     data->list_time, data->start, TRUE) > 0) &&
	    (jana_utils_time_compare (
	     data->list_time, data->end, TRUE) < 0)) {
		g_object_unref (data->start);
		data->start = jana_time_duplicate (data->list_time);
	}
	jana_utils_time_set_start_of_week (data->start);
	g_object_unref (data->end);
	data->end = jana_time_duplicate (data->start);
	jana_time_set_day (data->end, jana_time_get_day (data->end) + 7);
	
	refresh_view (data);
}

static void
page_hidden (DatesData *data)
{
	if (data->week_time) g_object_unref (data->week_time);
	data->week_time = jana_time_duplicate (data->start);
}

static void
notify_visible_cb (GObject *gobject, GParamSpec *arg1, DatesData *data)
{
	if ((!hidden) && (!GTK_WIDGET_VISIBLE (gobject))) {
		hidden = TRUE;
		page_hidden (data);
	}
}

static gboolean
visibility_notify_event_cb (GtkWidget *widget, GdkEventVisibility *event,
			    DatesData *data)
{
	if (((event->state == GDK_VISIBILITY_PARTIAL) ||
	     (event->state == GDK_VISIBILITY_UNOBSCURED)) && (hidden)) {
		hidden = FALSE;
		page_shown (data);
	}/* else if ((event->state == GDK_VISIBILITY_FULLY_OBSCURED) &&
		   (!hidden)) {
		hidden = TRUE;
	}*/
	
	return FALSE;
}

static void
unmap_cb (GtkWidget *widget, DatesData *data)
{
	if (!hidden) {
		hidden = TRUE;
		page_hidden (data);
	}
}

static void
new_clicked_cb (GtkToolButton *button, DatesData *data)
{
	JanaTime *time;
	JanaDuration *selection = jana_gtk_day_view_get_selection (
		JANA_GTK_DAY_VIEW (data->week_view));
	
	if (selection)
		time = jana_time_duplicate (selection->start);
	else
		time = jana_time_duplicate (data->start);
	if (jana_time_get_isdate (time)) {
		jana_time_set_isdate (time, FALSE);
		jana_time_set_hours (time, 9);
	}
	
	dates_utils_create_new_event (data, time);
	
	if (selection) {
		jana_event_set_end (data->selected_event, selection->end);
		jana_duration_free (selection);
	}
	
	g_object_unref (time);
	
	data->editing = TRUE;
	gtk_notebook_set_current_page (GTK_NOTEBOOK (data->notebook), -1);
}

static void
back_clicked_cb (GtkToolButton *button, DatesData *data)
{
	jana_time_set_day (data->start, jana_time_get_day (data->start) - 7);
	jana_time_set_day (data->end, jana_time_get_day (data->end) - 7);
	refresh_view (data);
}

static void
home_clicked_cb (GtkToolButton *button, DatesData *data)
{
	g_object_unref (data->start);
	g_object_unref (data->end);
	
	data->start = jana_ecal_utils_time_today (data->location);
	jana_utils_time_set_start_of_week (data->start);
	data->end = jana_time_duplicate (data->start);
	jana_time_set_day (data->end, jana_time_get_day (data->end) + 7);
	
	refresh_view (data);
}

static void
forward_clicked_cb (GtkToolButton *button, DatesData *data)
{
	jana_time_set_day (data->start, jana_time_get_day (data->start) + 7);
	jana_time_set_day (data->end, jana_time_get_day (data->end) + 7);
	refresh_view (data);
}

static void
search_text_changed_cb (MokoSearchBar *bar, GtkEditable *editable,
			DatesData *data)
{
	if (hidden || (!moko_search_bar_search_visible (bar))) return;
	
	if (data->category) {
		g_free (data->category);
		data->category = NULL;
	}
	
	g_free (data->search);
	data->search = g_utf8_casefold (gtk_entry_get_text (
		GTK_ENTRY (editable)), -1);
}

static void
search_combo_changed_cb (MokoSearchBar *bar, GtkComboBox *combo_box,
			 DatesData *data)
{
	if (hidden || moko_search_bar_search_visible (bar)) return;
	
	if (data->search) {
		g_free (data->search);
		data->search = NULL;
	}
	
	g_free (data->category);
	if (gtk_combo_box_get_active (combo_box) > 0)
		data->category = gtk_combo_box_get_active_text (combo_box);
	else
		data->category = NULL;
}

static void
search_toggled_cb (MokoSearchBar *bar, gboolean search_visible,
		   DatesData *data)
{
	if (hidden) return;
	
	if (search_visible) {
		search_text_changed_cb (bar, (GtkEditable *)
			moko_search_bar_get_entry (bar), data);
	} else {
		search_combo_changed_cb (bar,
			moko_search_bar_get_combo_box (bar), data);
	}
}

static void
view_size_allocate_cb (GtkWidget *scroll, GtkAllocation *allocation,
		       GtkWidget *week_view)
{
	static gint last_height = 0;
	if (allocation->height != last_height) {
		last_height = allocation->height;
		gtk_widget_set_size_request (week_view, -1, last_height * 2);
	}
}

GtkWidget *
dates_create_week_page (DatesData *data)
{
	GtkWidget *vbox, *toolbar, *combo, *scroll, *viewport;
	GtkToolItem *button;
	
	vbox = gtk_vbox_new (FALSE, 0);
	
	/* Create toolbar */
	toolbar = gtk_toolbar_new ();
	gtk_box_pack_start (GTK_BOX (vbox), toolbar, FALSE, TRUE, 0);

	/* New button */
	button = dates_utils_toolbutton_new (GTK_STOCK_NEW);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), button, 0);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
		gtk_separator_tool_item_new (), 0);
	g_signal_connect (G_OBJECT (button), "clicked",
		G_CALLBACK (new_clicked_cb), data);

	/* Forward button */
	button = dates_utils_toolbutton_new (GTK_STOCK_GO_FORWARD);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), button, 0);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
		gtk_separator_tool_item_new (), 0);
	g_signal_connect (G_OBJECT (button), "clicked",
		G_CALLBACK (forward_clicked_cb), data);
	
	/* Home button */
	button = dates_utils_toolbutton_new (GTK_STOCK_HOME);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), button, 0);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
		gtk_separator_tool_item_new (), 0);
	g_signal_connect (G_OBJECT (button), "clicked",
		G_CALLBACK (home_clicked_cb), data);

	/* Back button */
	button = dates_utils_toolbutton_new (GTK_STOCK_GO_BACK);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), button, 0);
	g_signal_connect (G_OBJECT (button), "clicked",
		G_CALLBACK (back_clicked_cb), data);

	gtk_widget_show_all (toolbar);
	
	/* Create search box */
	combo = gtk_combo_box_new_text ();
	data->week_searchbar = moko_search_bar_new_with_combo (
		GTK_COMBO_BOX (combo));
	gtk_box_pack_start (GTK_BOX (vbox), data->week_searchbar,
		FALSE, TRUE, 0);
	gtk_widget_show (data->week_searchbar);
	
	/* Create week view */
	data->week_view = jana_gtk_day_view_new (NULL, 24);
	viewport = gtk_viewport_new (NULL, NULL);
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);
	gtk_container_add (GTK_CONTAINER (viewport), data->week_view);
	
	scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
		GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	gtk_container_add (GTK_CONTAINER (scroll), viewport);
	gtk_widget_show (data->week_view);
	gtk_widget_show (viewport);
	
	gtk_box_pack_start (GTK_BOX (vbox), scroll, TRUE, TRUE, 0);
	gtk_widget_show (scroll);
	
	g_signal_connect (scroll, "size-allocate",
		G_CALLBACK (view_size_allocate_cb), data->week_view);
	
	gtk_widget_add_events (data->week_view, GDK_VISIBILITY_NOTIFY_MASK);
	g_signal_connect (data->week_view, "visibility-notify-event",
		G_CALLBACK (visibility_notify_event_cb), data);
	g_signal_connect (data->week_view, "notify::visible",
		G_CALLBACK (notify_visible_cb), data);
	g_signal_connect (vbox, "unmap",
		G_CALLBACK (unmap_cb), data);
	
	g_signal_connect (data->week_searchbar, "text_changed",
		G_CALLBACK (search_text_changed_cb), data);
	g_signal_connect (data->week_searchbar, "combo_changed",
		G_CALLBACK (search_combo_changed_cb), data);
	g_signal_connect (data->week_searchbar, "toggled",
		G_CALLBACK (search_toggled_cb), data);

	return vbox;
}
